// Intrusive AVL

# include <stddef.h>
# include <stdlib.h>

# include "intrusive_avl.h"

// Rotations

static inline
void lr(struct node **t) {
  // Rotation
  struct node *s        = (*t)->right;
  (*t)->right           = s->left;
  s->left               = *t;
  *t                    = s;
  // Balance factor recomputing
  s->left->bal_factor   = -1 - s->bal_factor;
  s->bal_factor         =  - s->left->bal_factor;
}

static inline
void rr(struct node **t) {
  // Rotation
  struct node *s        = (*t)->left;
  (*t)->left            = s->right;
  s->right              = *t;
  *t                    = s;
  // Balance factor recomputing
  s->right->bal_factor  = 1 - s->bal_factor;
  s->bal_factor         =  - s->right->bal_factor;
}

static inline
void rlr(struct node **t) {
  // Rotation
  struct node *s        = (*t)->right->left;
  (*t)->right->left     = s->right;
  s->right              = (*t)->right;
  (*t)->right           = s->left;
  s->left               = *t;
  *t                    = s;
  // Balance factor recomputing
  s->left->bal_factor   = (s->bal_factor * (s->bal_factor - 1)) >> 1;
  s->right->bal_factor  = - (s-> bal_factor * (s->bal_factor + 1)) >> 1;
  s->bal_factor         = 0;
}

static inline
void lrr(struct node **t) {
  // Rotation
  struct node *s        = (*t)->left->right;
  (*t)->left->right     = s->left;
  s->left               = (*t)->left;
  (*t)->left            = s->right;
  s->right              = *t;
  *t                    = s;
  // Balance factor recomputing
  s->left->bal_factor   = (s->bal_factor * (s->bal_factor - 1)) >> 1;
  s->right->bal_factor  = - (s-> bal_factor * (s->bal_factor + 1)) >> 1;
  s->bal_factor         = 0;
}

// Find
static inline
struct node* find_(struct node *cur, cmp_fun_type cmp, struct node *x) {
  int r;
  while (cur && (r = cmp(x, cur))) {
    if (r < 0)
      cur = cur->left;
    else
      cur = cur->right;
  }
  return cur;
}

struct node* avl_find(struct intrusive_avl *tree, struct node *x) {
  return find_(tree->root, tree->cmp, x);
}

// Insert
static
int insert_(struct node **cur, cmp_fun_type cmp,
            struct node *new_node, struct node **inserted_node) {
  if (*cur == NULL) {
    // Insertion point
    *cur = new_node;
    new_node->bal_factor = 0;
    new_node->left = new_node->right = NULL;
    *inserted_node = new_node;
    return 1;
  }
  int c = cmp(new_node, *cur);
  if (c == 0) {
    // Key present
    *inserted_node = *cur;
    return -42;
  }
  struct node         **target;
  int                   dir, tmp;
  if (c < 0) {
    target = &((*cur)->left);
    dir = 1;
  } else {
    target = &((*cur)->right);
    dir = -1;
  }

  tmp = insert_(target, cmp, new_node, inserted_node);
  if ( tmp == 0 || tmp == -42)
    return tmp;
  (*cur)->bal_factor += dir;

  switch ((*cur)->bal_factor) {
  case -2:
    if ((*cur)->right->bal_factor == 1) rlr(cur);
    else lr(cur);
    break;
  case 2:
    if ((*cur)->left->bal_factor == -1) lrr(cur);
    else rr(cur);
    break;
  }

  return (*cur)->bal_factor;
}

int avl_insert(struct intrusive_avl *tree,
               struct node *new_node, struct node **inserted_node) {
  return insert_(&(tree->root), tree->cmp, new_node, inserted_node) != -42;
}

// Delete

static
int del_min(struct node **cur, struct node **target) {
  if ((*cur)->left == NULL) {
    *target = *cur;
    *cur = (*target)->right;
    return 1;
  }
  if ( del_min(&((*cur)->left), target) == 0)
    return 0;
  int bal = ( (*cur)->bal_factor -= 1 );
  if (bal == -2) {
    if ((*cur)->right->bal_factor == 1)
      rlr(cur);
    else
      lr(cur);
    bal = (*cur)->bal_factor;
  }
  return 1 - bal * bal;
}

static
int del_max(struct node **cur, struct node **target) {
  if ((*cur)->right == NULL) {
    *target = *cur;
    *cur = (*target)->right;
    return 1;
  }
  if ( del_max(&((*cur)->right), target) == 0)
    return 0;
  int bal = ( (*cur)->bal_factor -= 1 );
  if (bal == 2) {
    if ((*cur)->left->bal_factor == -1)
      lrr(cur);
    else
      rr(cur);
    bal = (*cur)->bal_factor;
  }
  return 1 - bal * bal;
}

static inline
void detach_node(struct node *b) {
  b->left = b->right = NULL;
  b->bal_factor = 0;
}

static inline
void node_swap(struct node *a, struct node *b) {
  a->left = b->left;
  a->right = b->right;
  a->bal_factor = b->bal_factor;
  detach_node(b);
}

static
int delete_(struct node **cur, cmp_fun_type cmp,
            struct node *x, struct node **target) {
  if (*cur == NULL) {
    *target = NULL;
    return 0;
  }
  int r = cmp(x, *cur);
  if (r == 0) {
    *target = *cur;
    if ((*cur)->left && (*cur)->right) {
      if ((*cur)->bal_factor < 0) {
        int b = del_min(&((*cur)->right), cur);
        node_swap(*cur, *target);
        if ( b == 0 ) return 0;
        (*cur)->bal_factor = 0;
        return 1;
      } else {
        int b = del_max(&((*cur)->left), cur);
        node_swap(*cur, *target);
        if ( b == 0 ) return 0;
        int bal = ( (*cur)->bal_factor -= 1 );
        return 1 - (bal * bal);
      }
    }
    if ((*cur)->left != NULL)
      *cur = (*cur)->left;
    else
      *cur = (*cur)->right;
    detach_node(*target);
    return 1;
  }

  int bal;
  if (r < 0) {
    if (delete_(&((*cur)->left), cmp, x, target) == 0)
      return 0;
    bal = (*cur)->bal_factor -= 1;
    if (bal == -2) {
      if ((*cur)->right->bal_factor == 1)
        rlr(cur);
      else
        lr(cur);
      bal = (*cur)->bal_factor;
    }
  } else {
    if (delete_(&((*cur)->right), cmp, x, target) == 0)
      return 0;
    bal = (*cur)->bal_factor += 1;
    if (bal == 2) {
      if ((*cur)->left->bal_factor == -1)
        lrr(cur);
      else
        rr(cur);
      bal = (*cur)->bal_factor;
    }
  }
  return 1 - bal * bal;
}

struct node* avl_delete(struct intrusive_avl *tree, struct node *x) {
  struct node          *target = NULL;
  delete_(&(tree->root), tree->cmp, x, &target);
  return target;
}

// Clear tree
static
void reset_(struct node **cur, void (*free_node)(struct node*)) {
  if (*cur) {
    reset_(&((*cur)->left), free_node);
    reset_(&((*cur)->right), free_node);
    free_node(*cur);
    *cur = NULL;
  }
}

void avl_reset(struct intrusive_avl *tree, void (*free_node)(struct node*)) {
  reset_(&(tree->root), free_node);
}
